package com.googlecode.marrowboy;

import java.io.FileNotFoundException;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.List;

import org.hamcrest.Description;
import org.hamcrest.Matcher;
import org.hamcrest.MatcherAssert;
import org.hamcrest.StringDescription;

import com.googlecode.marrowboy.assertionresult.AssertionResult;
import com.googlecode.marrowboy.assertionresult.TestRunResult;
import com.googlecode.marrowboy.writers.OutputConfiguration;
import com.googlecode.marrowboy.writers.ReportWriter;
import com.googlecode.marrowboy.writers.Templates;
import com.googlecode.marrowboy.writers.VelocityTemplateWriter;

/**
 * Provides methods which can be used in place of the Hamcrest methods, and
 * methods to configure the behaviour of MarrowBoy reports.
 * 
 * This class contains the only methods which are expected to be called by
 * users. It is expected to be used statically so instantiation is forbidden.
 * 
 * @author mjg
 */
public final class MarrowBoyAssert {

	private static final TestRunResult messages = new TestRunResult();

	private static final String UNSPECIFIED_REASON = "";

	private static List<OutputConfiguration> outputConfigurations = new ArrayList<OutputConfiguration>();

	static {

		try {

			// TODO: Don't hard-code directory name here
			final String reportFilename = "target/marrowboy-report.html";

			outputConfigurations.add(
					new OutputConfiguration(
							new VelocityTemplateWriter(Templates.DEFAULT_HTML_TEMPLATE),
							new PrintStream(reportFilename)));

		} catch (final FileNotFoundException e) {
			e.printStackTrace();
			throw new RuntimeException(e);
		}

		Runtime.getRuntime().addShutdownHook(new Thread() {

			@Override
			public void run() {

				for (final OutputConfiguration thisConfig : outputConfigurations) {

					if (thisConfig.getReportWriter() != null && thisConfig.getPrintStream() != null) {
						thisConfig.getReportWriter().writeReport(messages, thisConfig.getPrintStream());
					}

				}

			}

		});

	}

	/**
	 * No instantiation.
	 */
	private MarrowBoyAssert() {
	}

	/**
	 * Allows users to specify the {@link ReportWriter} to use (eg
	 * {@link PlainTextWriter}), and the {@link PrintStream} to write to.
	 * 
	 * If either of these is null, the report will not be written anywhere. If
	 * this method is called more than once, the last call will overwrite all
	 * previous calls. If it is never called, we use a {@link PlainTextWriter}
	 * to {@link System#out}.
	 * 
	 * @param templateName
	 *            the template to use. MarrowBoy will attempt to find a file
	 *            with this name on the classpath.
	 * 
	 * @param printStream
	 *            the {@link PrintStream} to write to. Easy options for this
	 *            argument are {@link System#out} or
	 *            {@link PrintStream#PrintStream(String)}
	 */
	public static void addReportTemplate(final String templateName, final PrintStream printStream) {
		outputConfigurations.add(new OutputConfiguration(new VelocityTemplateWriter(templateName), printStream));
	}

	/**
	 * Remove all {@link OutputConfiguration}s - maybe useful because the
	 * default is to include {@link PlainTextWriter} to {@link System#out}
	 */
	public static void clearReportWriters() {
		outputConfigurations.clear();
	}

	/**
	 * Provides the same functionality as
	 * {@link MatcherAssert#assertThat(Object, Matcher)}, except that it also
	 * adds the assertion to the generated report.
	 * 
	 * @param actual
	 *            the object produced by the system under test.
	 * @param matcher
	 *            the {@link Matcher} which will be applied to the first
	 *            argument.
	 * @return the {@link AssertionResult} which was generated for this
	 *         assertion.
	 */
	public static <T> AssertionResult assertThat(final T actual, final Matcher<? super T> matcher) {

		try {
			ProxyAssert.assertThat(actual, matcher);

		} catch (final AssertionError ae) {
			realLogThat(UNSPECIFIED_REASON, actual, matcher, false);
			throw ae;

		}

		return realLogThat(UNSPECIFIED_REASON, actual, matcher, true);

	}

	/**
	 * Provides the same functionality as
	 * {@link MatcherAssert#assertThat(String, Object, Matcher)}, except that it
	 * also adds the assertion to the generated report.
	 * 
	 * @param reason
	 *            some textual description of the assertion.
	 * @param actual
	 *            the object produced by the system under test.
	 * @param matcher
	 *            the {@link Matcher} which will be applied to the first
	 *            argument.
	 * @return the {@link AssertionResult} which was generated for this
	 *         assertion.
	 */
	public static <T> AssertionResult assertThat(final String reason, final T actual, final Matcher<? super T> matcher) {

		try {
			ProxyAssert.assertThat(actual, matcher);

		} catch (final AssertionError ae) {
			realLogThat(reason, actual, matcher, false);
			throw ae;

		}

		return realLogThat(reason, actual, matcher, true);

	}

	private static <T> AssertionResult realLogThat(final String reason, final T actual, final Matcher<? super T> matcher, final boolean wasSuccessful) {

		// if we pop this many elements off the call stack, we'll be looking at
		// the method which first called us.
		final int stackPopAmount = 3;

		StackTraceElement stackTraceElement = Thread.currentThread().getStackTrace()[stackPopAmount];

		// This happens if we're using the java agent, because of an extra
		// bytecode-generated proxy call
		// TODO: Do this better...
		if ("org.hamcrest.MatcherAssert".equals(stackTraceElement.getClassName())) {
			stackTraceElement = Thread.currentThread().getStackTrace()[stackPopAmount + 1];
		}

		final String actualString;

		if (actual == null) {
			actualString = "null";
		} else {
			actualString = actual.toString();
		}

		// TODO: try to do other types nicely here

		final Description description = new StringDescription();
		matcher.describeTo(description);

		return messages.logMessageFromStackTraceElement(stackTraceElement, reason, actualString, description.toString(), wasSuccessful);

	}

}
