package com.googlecode.marrowboy.utilities;

/**
 * Various methods to help when dealing with {@link String} objects.
 * 
 * @author mjg
 */
public class StringHelper {

	/**
	 * Constructor.
	 */
	public StringHelper() {
	}

	/**
	 * Turns a fully-qualified class name into the simple name. <br/>
	 * eg: "java.lang.String" -> "String"
	 * 
	 * @param fqName
	 *            the fully-qualified class name.
	 * @return the simple class name.
	 */
	public String stripClassName(final String fqName) {
		return fqName.substring(fqName.lastIndexOf(".") + 1);
	}

	/**
	 * Turns a camelCase String into a String in sentence case, with spaces. <br/>
	 * eg "aMethodName" -> "A method name"<br/>
	 * eg "AClassName" -> "A class name"
	 * 
	 * @param ccString
	 *            a string in camelCase
	 * @return the string in sentence case
	 */
	public String camelCaseToSentence(final String ccString) {

		final StringBuilder out = new StringBuilder();

		boolean firstChar = true;

		final char[] camelCaseCharArray = ccString.toCharArray();

		for (int i = 0; i < camelCaseCharArray.length; i++) {
			final char c = camelCaseCharArray[i];

			if (firstChar) {
				out.append(Character.toUpperCase(c));
				firstChar = false;

			} else if (Character.isUpperCase(c)) {

				if (i == camelCaseCharArray.length - 1) {
					// final character is uppercase
					if (Character.isUpperCase(camelCaseCharArray[i - 1])) {
						out.append(c);
					} else {
						out.append(" ").append(c);
					}

				} else if (Character.isLowerCase(camelCaseCharArray[i + 1])) {
					// we started a new word
					out.append(" ").append(Character.toLowerCase(c));

				} else if (Character.isLowerCase(camelCaseCharArray[i - 1])) {
					// start of an acronym?

					out.append(" ");

					if (i + 2 >= camelCaseCharArray.length || Character.isUpperCase(camelCaseCharArray[i + 2])) {
						// yes, an acronym
						out.append(c);

					} else {
						// no, not an acronym any more
						out.append(Character.toLowerCase(c));

					}

				} else {
					out.append(c);

				}

			} else {
				out.append(c);
			}

		}

		return out.toString();

	}

	/**
	 * Retrieve the class name in sentence case from a {@link StackTraceElement}
	 * 
	 * @param stackTraceElement
	 *            the {@link StackTraceElement}
	 * @return the class name, in sentence case
	 */
	public String extractClassName(final StackTraceElement stackTraceElement) {
		final String simpleName = stripClassName(stackTraceElement.getClassName());
		return camelCaseToSentence(simpleName);
	}

	/**
	 * Retrieve the method name in sentence case from a
	 * {@link StackTraceElement}
	 * 
	 * @param stackTraceElement
	 *            the {@link StackTraceElement}
	 * @return the method name, in sentence case
	 */
	public String extractMethodName(final StackTraceElement stackTraceElement) {
		return camelCaseToSentence(stackTraceElement.getMethodName());
	}

}
